require 'eventmachine'
require 'active_support/all'
require 'nginx_controller/docker/listener'
require 'nginx_controller/docker/query'
require 'nginx_controller/docker/service'
require 'nginx_controller/master_service'
require 'yaml'


module NginxController
  def self.ssl_root
    ENV['SSL_ROOT'].presence || '/etc/ssl/proxy'
  end

  def self.certs_path
    ENV['SSL_CERTS_ROOT'].presence || File.join(ssl_root, 'certs')
  end

  def self.htpasswd_path
    ENV['HTPASSWD_ROOT'].presence || '/etc/nginx/htpasswd'
  end

  class Controller
    def initialize
      @services = []

      @reload = {}
      @watchlist = {}
      @le_queue = {}
    end

    def run
      EM.run {
        init
        Docker::Listener.run do | _ |
          update_services_status
        end

        # when a service changes from starting to running, no event is raised
        # add all new services to a watchlist and poll docker until they are running or until a 120s timeout is reached
        EventMachine::PeriodicTimer.new(5) do
          if @watchlist.keys.any?
            update_services_status
          end
        end

        EventMachine::PeriodicTimer.new(5) do
          process_letsencrypt_queue
        end

        every_day_at_midnight do
          check_letsencrypt_status
        end
      }
    end

    def init
      services_ps do | services |
        update_services(services)
        @services = services
        STDOUT.flush
      end
    end

    protected

    def update_services_status
      services_ps do | services |
        update_services(services)
        @services = services
        STDOUT.flush
      end
    end

    def check_letsencrypt_status
      services_ps do | services |
        puts '-> Check expiring certificates'
        services.each do | svc |
          le_queue_add(svc) if svc.letsencrypt_expiring?
          STDOUT.flush
        end
        STDOUT.flush
      end
    end

    def process_letsencrypt_queue
      # process le only when the rervices are settled
      if @watchlist.blank?
        id, svc = @le_queue.first
        if svc
          puts "-> Update letsencrypt for #{svc.name}"
          le_queue_remove(svc)
          result = svc.letsencrypt_request!
          reload_service(svc) if result
          STDOUT.flush
        end
      end
    end

    def update_services(services)
      watchlist_timeout

      added = list_new_services(services)
      changed = list_changed_services(services)
      removed = list_removed_services(services)
      cleanup = []
      errors = []

      removed.each do | svc |
        puts "Deleted service #{svc.name}"
        begin
          svc.deconfigure!
        rescue
          puts "Cannot deactivate service #{svc.name}: #{$!}"
          puts "\t" + $!.backtrace[0..10].join("\n\t")
          errors << $!
        ensure
          watchlist_remove(svc)
        end
      end

      added.each do | svc |
        puts "New service #{svc.name}"
        begin
          svc.configure!
          watchlist_add(svc) unless svc.settled?
          le_queue_add(svc) if svc.letsencrypt_invalid?
        rescue
          puts "Cannot activate service #{svc.name}: #{$!}"
          puts "\t" + $!.backtrace[0..10].join("\n\t")
          errors << $!
        end
      end

      changed.each do | pair |
        puts "Updated service #{pair[1].name}"
        print_diff(pair[0].diff(pair[1]))
        begin
          pair[1].reconfigure!
          if pair[1].settled?
            watchlist_remove(pair[1])
          else
            watchlist_add(pair[1])
          end
        rescue
          watchlist_remove(pair[1])
          puts "Cannot activate service #{pair[1].name}: #{$!}"
          puts "\t" + $!.backtrace[0..10].join("\n\t")
          errors << $!
        end
      end

      cleanup = cleanup_config_files(services)

      if added.any? || changed.any? || removed.any? || cleanup.any?
        puts "Reloading nginx..."
        out = `nginx -s reload 2>&1`
        puts out
      end
    end

    # query docker for all services and run the block with the list of Service objects
    def services_ps(&block)
      Docker::Query.exec(:services).callback do | svcs |

        ids = svcs.inject({}) do | memo, svc |
          memo[svc['ID']] = true
          memo
        end

        Docker::Query.exec(:tasks).callback do | res |
          tasks = res.inject({}) do | memo, t |
            if t.is_a?(Hash)
              memo[t['ServiceID']] ||= []
              memo[t['ServiceID']] << t
            else
              puts "WARN: unexpected task #{t.inspect}"
            end
            memo
          end

          services = svcs.inject([]) do | memo, svc |
            id = svc['ID']
            svc = NginxController::Docker::Service.new(svc, tasks[id])
            memo << svc if svc.managed?
            memo
          end

          services.unshift MasterService.new
          services.sort!

          block.call services
        end
      end
    end

    def list_new_services(services)
      ids = services.map(&:id)
      cur = @services.map(&:id)
      ids = ids - cur
      services.select{|s| s.id.in?(ids)}
    end

    # extract all updated services as pairs [old, new]
    def list_changed_services(services)
      newSvcs = services.inject({}){|m,s| m[s.id]=s; m}
      oldSvcs = @services.inject({}){|m,s| m[s.id]=s; m}
      ids = newSvcs.keys
      cur = oldSvcs.keys
      ids = ids & cur
      ids.inject([]) do | memo, id |
        newSvc = newSvcs[id]
        oldSvc = oldSvcs[id]
        changed = !(oldSvc === newSvc)
        memo << [oldSvc, newSvc] if changed || @reload[id]
        @reload.delete(id)
        memo
      end
    end

    def list_removed_services(services)
      ids = services.map(&:id)
      cur = @services.map(&:id)
      ids = cur - ids
      @services.select{|s| s.id.in?(ids)}
    end

    def cleanup_config_files(active_services)
      configs = active_services.map(&:config_filename)
      leftovers = Dir['/etc/nginx/vhosts.d/*.conf'].inject([]) do | memo, cfg |
        unless cfg.in?(configs)
          memo << cfg
          puts "Cleanup old config file #{cfg}"
          FileUtils.rm_f(cfg)
        end
        memo
      end
      leftovers
    end

    def watchlist_add(svc)
      unless @watchlist.keys.include?(svc.id)
        @watchlist[svc.id] = [svc, 0]
      end
    end

    def watchlist_remove(svc)
      @watchlist.delete(svc.id)
    end

    def watchlist_timeout
      @watchlist = @watchlist.inject({}) do | memo, pair |
        id, svc = pair
        svc[1] += 5
        memo[id] = svc if svc[1] < 120
        memo
      end
    end

    def le_queue_add(svc)
      unless @le_queue[svc.id]
        puts "Service #{svc.name} should update letsencrypt"
        @le_queue[svc.id] = svc
      end
    end

    def le_queue_remove(svc)
      @le_queue.delete(svc.id)
    end

    def reload_service(svc)
      @reload[svc.id] = svc
      watchlist_add(svc)
    end

    def print_diff(diff)
      str = diff.map do | k, d |
        if k == :labels
          ["  labels: "] + d.map do | l, ld |
            "    #{l}: #{ld[0] || 'nil' } => #{ld[1]}"
          end
        else
          "  #{k}: #{d[0] || 'nil' } => #{d[1]}"
        end
      end
      puts str.flatten.join("\n")
    end

    def every_day_at_midnight(&block)
      run_at = (1.days.from_now.at_midnight - Time.new).to_i + 5
      EventMachine::Timer.new(run_at) do
        block.call
        every_day_at_midnight(&block)
      end
    end

  end
end
