gem 'acme-client'
gem 'activesupport'
require 'nginx_controller/acme/base'
require 'fileutils'
require 'openssl'
require 'acme-client'
require 'uri'
require 'active_support/all'
module NginxController
  module Acme
    class Ruby < NginxController::Acme::Base
      require 'nginx_controller/acme/ruby/order'

      def self.server_private_key
        @server_pkey ||= begin
          pkey = nil
          pkey_file = File.join(NginxController::Acme.le_root, "server_private_key.rsa")
          unless File.exists?(pkey_file)
            puts "  - Generating server private key #{pkey_file}"
            FileUtils.mkdir_p(NginxController::Acme.le_root)
            pkey = OpenSSL::PKey::RSA.new(4096)
            File.open(pkey_file, 'w+') do | f |
              f.write pkey
            end
          end
          pkey || begin
            puts "  - Reading server private key #{pkey_file}"
            OpenSSL::PKey::RSA.new File.read(pkey_file)
          end
        end
      end

      def valid?
        live_privkey? && live_cert? && expires_at > Time.now
      end

      def expiring?
        live_privkey? && live_cert? && expires_at.to_date < 30.days.from_now.to_date
      end

      def has_names?
        options[:names].find{|name| !name.in?(names) }.blank?
      end

      def expires_at
        live_cert.try(:not_after) || Time.at(0)
      end

      def privkey_path
        live_privkey_path
      end
  
      def cert_path
        live_fullchain_path
      end 

      def names
        names = []
        if live_cert?
          name = live_cert.subject.to_a.find{|x| x[0] == 'CN'}
          names << name[1] if name

          ext = live_cert.extensions.find{|x| x.oid == 'subjectAltName'}
          names += ext.value.split(',').map{|n| n.strip.gsub(/\Adns:/i, '') } if ext
          names.uniq!
        end
        names
      end

      def root_path
        File::join(NginxController::Acme.le_root, @options[:service_name])
      end

      def live_cert_dir
        File::join(root_path, 'current')
      end

      def live_privkey_path
        File::join(live_cert_dir, 'key.pem')
      end

      def live_privkey?
        File.exists?(live_privkey_path)
      end

      def live_privkey
        @live_privkey ||= OpenSSL::PKey.read(File.read(live_privkey_path))
      end

      def live_cert_path
        File.join(live_cert_dir, 'cert.pem')
      end

      def live_cert?
        File.exists?(live_cert_path)
      end

      def live_cert
        @live_cert ||= OpenSSL::X509::Certificate.new(File.read(live_cert_path))
      end

      def live_chain_path
        File.join(live_cert_dir, 'chain.pem')
      end

      def live_fullchain_path
        File::join(live_cert_dir, 'full_chain.pem')
      end

      def prepare!
        begin
          do_prepare
        rescue NginxController::Acme::Ruby::Order::PartialAuthFailure => ex
          puts "-- PartialAuthFailure ----------------"
          puts ex.message
          puts "--------------------------------------"
          do_prepare(@order.autorized_names)
          raise ex
        end
      end

      def authorized?
        @authorized
      end

      def request_certificate!
        @certs = nil

        if @authorized && !@cert_requested
          puts "  - Build certificate for #{@order.autorized_names}"
          csr = ::Acme::Client::CertificateRequest.new(
            names: @order.autorized_names,
            private_key: self.private_key
          )
          order = @order.order
          order.finalize(csr: csr)
          sleep(1) while order.status == 'processing'

          begin
            fullchain = order.certificate
            certs = split_chain fullchain
            backup!
            @certs = save_certs!([
              self.private_key.to_pem,
              certs[:cert],
              certs[:chain],
              fullchain
            ])

          #  @certs = []
          #  @certs[0] = cert.request.private_key.to_pem
          #  @certs[1] = cert.to_pem
          #  @certs[2] = cert.chain_to_pem
          #  @certs[3] = cert.fullchain_to_pem
          rescue
            puts "   ! Cannot generate SSL cert: #{$!}"
            @certs = []
          end
          @cert_requested = true
        end
        @certs
      end

      protected

      def client
        @client ||= begin
          reg_file = File::join(NginxController::Acme.le_root, NginxController::Acme.acme_contact)

          dir = NginxController::Acme.acme_endpoint.strip

          kid = begin
            if File::exists?(reg_file)
              File.read(reg_file).strip
            end
          end

          client_options = {
            private_key: self.class.server_private_key,
            directory: dir,
            connection_options: { request: { open_timeout: 5, timeout: 5 } }
          }

          client_options[:kid] = kid unless kid.blank?

          client = ::Acme::Client.new(client_options)

          if kid.blank?
            begin
              puts "   - Register contact #{NginxController::Acme.acme_contact}"
              registration = client.new_account(contact: "mailto:#{NginxController::Acme.acme_contact}", terms_of_service_agreed: true)
              FileUtils.touch(reg_file)
              File.open(reg_file, 'w+') do | f |
                f.write registration.kid
              end
            rescue ::Acme::Client::Error::Malformed
              if $!.to_s =~ /already in use/
                FileUtils.touch(reg_file)
              else
                raise $!
              end
            end
          end
          client
        end
      end

      def private_key
        @pkey ||= begin
          begin
            pkey = live_privkey
          rescue
            puts "Cannot load private key #{live_privkey_path}: #{$!}"
          end
          if !pkey
            puts "   - Generating private key"
            pkey = OpenSSL::PKey::RSA.new(4096)
          end
          pkey
        end
      end

      def do_prepare(names = @options[:names])
        unless @prepared
          @authorized ||= begin
            puts " - Prepare Let's Encrypt"
            #@domains = [@service.servername, @service.aliases].flatten.compact.map do | dom |
            #  dom = Domain.new(dom, client, @service)
            #end
            
            @order = Order.new(names, client)
            @order.authorize! 

            @order.authorized?            
          end
          @prepared = true
        end
        @authorized
      end

      def split_chain(pem)
        names = @options[:names]
        certs = {
          cert: nil,
          chain: []
        }
        parts = pem.gsub(/^-+(BEGIN|END) CERTIFICATE-+$/, '-----').split('-----').map(&:strip).reject{|s| s == ''}
        parts.each do | part |
          cert = OpenSSL::X509::Certificate.new "-----BEGIN CERTIFICATE-----\n#{part}\n-----END CERTIFICATE-----"
          is_cert = names.find{|n| cert.subject.eql? OpenSSL::X509::Name.new([['CN', n]]) }.present?
          if is_cert
            certs[:cert] = "-----BEGIN CERTIFICATE-----\n#{part}\n-----END CERTIFICATE-----\n"
          else
            certs[:chain] << "-----BEGIN CERTIFICATE-----\n#{part}\n-----END CERTIFICATE-----\n"
          end
        end
        certs[:chain] = certs[:chain].join('')
        certs
      end

      def le_backup_cert_dir
        File::join(root_path, 'backup')
      end

      def backup!
        if File::exists?(live_cert_dir)
          puts "   - Backup \"#{live_cert_dir}\" to \"#{le_backup_cert_dir}\""
          if File::directory?(le_backup_cert_dir)
            FileUtils::remove_dir(le_backup_cert_dir)
          end
          FileUtils::move(live_cert_dir, le_backup_cert_dir)
        end
      end

      def save_certs!(certs)
        ret = []
        if certs.any?
          puts "   - Save certificate to \"#{live_cert_dir}\""
          FileUtils::mkdir_p(live_cert_dir)
          ret << (pem = File::join(live_cert_dir, 'key.pem'))
          File::open(pem, 'w+') do | f |
            f.write certs[0]
          end
          ret << (pem = File::join(live_cert_dir, 'cert.pem'))
          File::open(pem, 'w+') do | f |
            f.write certs[1]
          end
          ret << (pem = File::join(live_cert_dir, 'chain.pem'))
          File::open(pem, 'w+') do | f |
            f.write certs[2]
          end
          ret << (pem = File::join(live_cert_dir, 'full_chain.pem'))
          File::open(pem, 'w+') do | f |
            f.write certs[3]
          end
        end
        ret
      end
    end
  end
end
