#gem 'net-dns', '~> 0.8'
require 'open3'
#require 'inifile'

module NginxController
  module Acme
    class Ruby
      class Order
        class BaseAuthFailure < StandardError
          def failed_challenges(challenges)
            challenges.inject([]) do | m, ch |
              err = ch[1].error
              err = err['detail'] if err.is_a?(Hash)
              m << "- #{ch[0].domain}: #{err}" if ch[1].status != 'valid'
              m
            end
          end

          def initialize(msg, challenges)
            msg += failed_challenges(challenges)
            super msg.join("\n")
          end
        end

        class AuthFailure < BaseAuthFailure
          def initialize(challenges)
            super ['Fatal domain authorization failure', '', 'challenge authorization failed for:'], challenges
          end
        end

        class PartialAuthFailure < BaseAuthFailure
          def initialize(challenges)
            super ['Partial domain authorization failure', '', 'challenge authorization failed for:'], challenges
          end
        end

        LE_WELL_KNOWN_DIR = "/srv/validation/well-known"

        attr_reader :domains, :order
        def initialize(domains, acme_client)
          @domains = domains
          @client = acme_client
        end

        def acme_challenge
          'HTTP01'
        end

        def authorize!
          create_order!
          valid = true
          @challenges ||= order.authorizations.map do | auth |
            challenge = begin
              case acme_challenge.downcase.to_sym
              #when :dns01 then
              #  ch = auth.dns
              #  if prepare_dns_challenge!(auth, ch)
              #    ch
              #  else
              #    cleanup_dns_challenge!(auth, ch)
              #    valid = false
              #  end
              when :http01 then
                ch = auth.http
                if prepare_http_challenge!(auth, ch)
                  ch
                else
                  cleanup_http_challenge!(auth, ch)
                  valid = false
                end
              else
                puts "   --- unhandle acme challenge #{acme_challenge}"
                false
              end
            end
            [auth, challenge]
          end

          if valid
            #if service.use_dns01?
            #  puts "   - Wait for DNS propagation..."
            #  STDOUT.flush
            #  sleep 60
            #end

            @challenges.each do | challenge |
              if challenge[1].status != 'valid'
                challenge[1].request_validation
              end
            end

            # wait all authorizations

            todo = @challenges
            timeout = Time.new + 2.minutes
            puts "   - Waiting for all challenges..."
            STDOUT.flush
            while (todo.length > 0) && (timeout > Time.new)
              todo = todo.reject do | challenge |
                if challenge[1].status == 'pending'
                  sleep 5
                  challenge[1].reload
                  STDOUT.flush
                  false
                else
                  dom = challenge[0].domain
                  dom = "*.#{dom}" if challenge[0].wildcard
                  if challenge[1].status != 'valid'
                    puts "   - Challenge for #{dom} failed: #{challenge[1].status} #{challenge[1].error['detail']}"
                  else
                    puts "   - Challenge for #{dom}: success"
                  end
                  STDOUT.flush
                  true
                end
              end
            end
            STDOUT.flush
          end

          # cleanup

          @challenges.each do | challenge |
            if challenge[1]
              case acme_challenge.downcase.to_sym
              #when :dns01 then
              #  cleanup_dns_challenge! challenge[0], challenge[1]
              when :http01 then
                cleanup_http_challenge! challenge[0], challenge[1]
              else
                raise "unhandle acme challenge #{acme_challenge}"
              end
            end
          end

          STDOUT.flush

          raise AuthFailure.new(@challenges) if @challenges.count{|ch| ch[1].status != 'valid'} == @challenges.count
          
          @authorized = true
          raise PartialAuthFailure.new(@challenges) unless @challenges.count{|ch| ch[1].status == 'valid'} == @challenges.count
        end

        def authorized?
          @authorized
        end

        def autorized_names
          @challenges.inject([]) do | memo, pair |
            auth, challenge = pair
            memo << auth.domain if challenge.status == 'valid'
            memo
          end
        end

        protected

        def create_order!
          @order ||= begin
            puts "  - Create order for #{@domains}"
            @client.new_order(identifiers: @domains)
          end
        end

        def prepare_http_challenge!(authorization, challenge)
          challenge_filename = http_challenge_filename_for(challenge)
          dom = authorization.domain
          puts "   - Witing HTTP01 challenge for #{dom} to #{challenge_filename}"
          FileUtils::mkdir_p(File::dirname(challenge_filename))
          #FileUtils::chown(Proxy::Provision.config.username, Proxy::Provision.config.group, File::dirname(challenge_filename))
          File::open(challenge_filename, 'w+') do | f |
            f.write challenge.file_content
          end
          #FileUtils::chown(Proxy::Provision.config.username, Proxy::Provision.config.group, challenge_filename)
          true
        end

        def cleanup_http_challenge!(authorization, challenge)
          fname = http_challenge_filename_for(challenge)
          if File::exists?(fname)
            dom = authorization.domain
            puts "   - Cleanup HTTP01 challenge for #{dom}"
            File::unlink(fname)
          end
        end

        def http_challenge_filename_for(challenge)
          fname = challenge.filename
          fname = fname.split('/')
          fname.shift
          File::join(*[LE_WELL_KNOWN_DIR, fname].flatten)
        end

        #def prepare_dns_challenge!(authorization, challenge)
        #  status = false
        #  dom = authorization.domain
        #  dom = "*.#{dom}" if authorization.wildcard
        #  puts "   - Prepare DNS01 challenge for #{dom}"
        #
        #  commands = <<-EOC
        #    server #{dnssec_key[:server]}
        #    update add #{challenge.record_name}.#{authorization.domain} 30 #{challenge.record_type} #{challenge.record_content}
        #    send
        #  EOC
        #
        #  creds = "#{dnssec_key[:algorithm].to_s.downcase}:#{dnssec_key[:name]}:#{dnssec_key[:secret]}"
        #  nsupdate = "#{Proxy::Provision.config.nsupdate_path} -y '#{creds}'"
        #
        #  begin
        #    stdout, stderr, status = Open3.capture3(nsupdate, stdin_data: commands)
        #  rescue
        #    stderr = $!.to_s
        #    status = false
        #  end
        #
        #  unless status && status.success?
        #    puts "   --- DNS update failed!"
        #    puts "   --- #{stderr}"
        #  end
        #
        #  status
        #end
        
        #def cleanup_dns_challenge!(authorization, challenge)
        #  status = false
        #  dom = authorization.domain
        #  dom = "*.#{dom}" if authorization.wildcard
        #  puts "   - Cleanup DNS01 challenge for #{dom}"
        #
        #  commands = <<-EOC
        #    server #{dnssec_key[:server]}
        #    update delete #{challenge.record_name}.#{authorization.domain} 30 #{challenge.record_type}
        #    send
        #  EOC
        #
        #  creds = "#{dnssec_key[:algorithm].to_s.downcase}:#{dnssec_key[:name]}:#{dnssec_key[:secret]}"
        #  nsupdate = "#{Proxy::Provision.config.nsupdate_path} -y '#{creds}'"
        #
        #  begin
        #    stdout, stderr, status = Open3.capture3(nsupdate, stdin_data: commands)
        #  rescue
        #    stderr = $!.to_s
        #    status = false
        #  end
        #
        #  unless status && status.success?
        #    puts "   --- DNS delete failed!"
        #    puts "   --- #{stderr}"
        #  end
        #
        #  status
        #end
        
        #def dnssec_key
        #  @dnssec_key ||= begin
        #    begin
        #      ini = IniFile.load(service.dnssec_key)
        #      {
        #        server:    ini['global']['dns_rfc2136_server'],
        #        algorithm: ini['global']['dns_rfc2136_algorithm'],
        #        name:      ini['global']['dns_rfc2136_name'],
        #        secret:    ini['global']['dns_rfc2136_secret'],
        #      }
        #    rescue
        #      puts "    --- cannot load dnssec key #{service.dnssec_key}"
        #      puts "    --- #{$!.inspect}"
        #      {}
        #    end
        #  end
        #end
      end
    end
  end
end
