gem 'mail'
gem 'activesupport'
require 'active_support/all'
require 'mail'

module NginxController
  class BaseService

    def initialize
    end

    def name
      raise 'unimplemented'
    end

    def managed?
      raise 'unimplemented'
    end

    def running?
      raise 'unimplemented'
    end

    def settled?
      raise 'unimplemented'
    end

    def ==(other)
      raise 'unimplemented'
    end

    def ===(other)
      raise 'unimplemented'
    end

    def <=>(other)
      raise 'unimplemented'
    end

    def diff(other)
      raise 'unimplemented'
    end

    def config_filename
      File.join('/etc/nginx/vhosts.d/', "#{name}.conf")
    end

    def configure!
      config = render
      test_config!(config)
      write_config!(config)
    end

    def reconfigure!
      configure!
    end

    def deconfigure!
      if File.exists?(config_filename)
        FileUtils.rm(config_filename)
      end
    end

    def letsencrypt?
      false
    end

    def letsencrypt_invalid?
      if letsencrypt?
        !acme.valid? || !acme.has_names?
      end
    end

    def letsencrypt_expiring?
      if letsencrypt?
        acme.expiring?
      end
    end

    def letsencrypt_request!
      result = false
      if letsencrypt?
        begin
          begin
            acme.prepare!
          rescue NginxController::Acme::Ruby::Order::PartialAuthFailure => ex
            notify_failure ex
            # a partial auth failure is not a fatal error
            # the connector should reauthorize only the valid domains
          end
          if acme.authorized?
            result = !!acme.request_certificate!
          end
        rescue
          notify_failure $!
        end
      end
      result
    end

    protected
    
    def test_config!(config)
      if File.exists?(config_filename)
        FileUtils.mv(config_filename, "#{config_filename}.test-bck")
      end
      File.open(config_filename, 'wb+') {|f| f.write(config)}
      begin
        out = `nginx -t 2>&1`
        raise StandardError.new(out) unless $?.success?
      rescue
        FileUtils.mv(config_filename, "#{config_filename}.err")
        raise $!
      ensure
        if File.exists?("#{config_filename}.test-bck")
          FileUtils.mv("#{config_filename}.test-bck", config_filename)
        end
      end
    end

    def write_config!(config)
      File.open(config_filename, 'wb+') {|f| f.write(config)}
    end

    def render(template = 'default', locals = nil)
      raise 'unimplemented'
    end

    def get_binding
      binding
    end

    def notify_failure(ex)
      msg = ["ACME failure for servce #{name}", '---------------------------------------------------', '']
      if ex.is_a? NginxController::Acme::Ruby::Order::BaseAuthFailure
        msg << ex.message
      else
        msg << "#{ex.message} (#{ex.class.name})"
        msg += ex.backtrace.map{|l| "    #{l}"}
      end
      msg = msg.join("\n")

      puts msg

      notify_failure_via_mail(msg)
    end

    def notify_failure_via_mail(msg)
      begin
        if ENV['NOTIFICATION_RECIPIENTS'].present?
          mail = Mail.new do | mail |
            mail.from "nginx@#{ENV['DEFAULT_HOST']}"
            mail.to ENV['NOTIFICATION_RECIPIENTS'].split(',').map(&:strip)
            mail.subject "ACME failure for service #{name}"
            mail.body msg
          end

          if ENV['SMTP_SERVER'].presence
            options = {
              address: ENV['SMTP_SERVER'],
              port: (ENV['SMTP_PORT'].presence || 25).to_i,
            }
            options[:enable_starttls_auto] = options[:port] == 25 || options[:port] == 587
            options[:domain] = ENV['SMTP_DOMAIN'] if ENV['SMTP_DOMAIN'].present?
            options[:user_name] = ENV['SMTP_USERNAME'] if ENV['SMTP_USERNAME'].present?
            options[:password] = ENV['SMTP_PASSWORD'] if ENV['SMTP_PASSWORD'].present?

            mail.delivery_method :smtp, options
          end

          mail.deliver
        end
      rescue
        puts "### Email notification not sent: #{$!}"
      end
    end
  end
end