gem 'activesupport'
require 'json'
require 'erb'
require 'ostruct'
require 'date'
require 'fileutils'
require 'openssl'
require 'active_support/all'
require 'nginx_controller/base_service'
require 'nginx_controller/self_signed_cert'
require 'nginx_controller/acme/ruby'

module NginxController
  module Docker
    class Service < NginxController::BaseService
      class Bindable < OpenStruct
        def get_binding
          binding
        end
      end

      attr_reader :id, :name, :instances, :labels, :data, :tasks, :locals

      def initialize(svc, tasks)
        @instances = 0
        @labels = {}
        @tasks = []
        if svc
          @id = svc['ID']
          @name = svc['Spec']['Name']
          @instances = svc['Spec']['Mode'] && svc['Spec']['Mode']['Replicated'] && svc['Spec']['Mode']['Replicated']['Replicas']
          @instances ||= 1
          @labels = svc['Spec']['Labels'].select{|k,v| k =~ /\Avhosts-ng/}
          @data = svc
        end
        if tasks
          @tasks = tasks
        end
      end

      def managed?
        self.labels.any?
      end

      def running_instances
        self.tasks.select{|t|
          t['DesiredState'] == 'running' && t['Status']['State'] == 'running'
        }.count
      end

      def running?
        self.instances > 0 && self.running_instances > 0
      end

      def settled?
        self.instances == self.running_instances
      end

      def ==(other)
        other.is_a?(Service) &&
          self.id == other.id
      end

      def ===(other)
        other.is_a?(Service) &&
          self.id == other.id &&
          self.instances == other.instances &&
          self.running_instances == other.running_instances &&
          self.labels == other.labels &&
          self.ssl_certs == other.ssl_certs
      end

      def <=>(other)
        self.name <=> other.name
      end

      def diff(other)
        diff = {}
        other = Service.new(nil, nil) unless other.is_a?(Service)
        diff[:id] = [self.id, other.id] if self.id != other.id
        diff[:instances] = [
          "%s/%s" % [self.running_instances, self.instances],
          "%s/%s" % [other.running_instances, other.instances]
        ] if self.instances != other.instances || self.running_instances != other.running_instances
        if self.labels != other.labels
          diff[:labels] = (self.labels.keys + other.labels.keys).uniq.map do | k |
            next if self.labels[k] == other.labels[k]
            [k, [self.labels[k], other.labels[k]]]
          end.compact
          diff[:labels] = Hash[diff[:labels]]
        end
        diff
      end

      def to_s
        "%s: %s/%s" % [self.name, self.instances, self.running_instances]
      end

      def rotate_logs?
        #labels['vhosts-ng/service/rotate-logs'] == "true"
        true
      end

      def logrotate_config_filename
        File.join('/etc/logrotate.d/', name)
      end

      def configure!
        create_directories
        super
        if rotate_logs?
          config = render_logrotate
          write_logrotate_config!(config)
        end
      end

      def deconfigure!
        super
        if File.exists?(logrotate_config_filename)
          FileUtils.rm_f(logrotate_config_filename)
        end
      end

      def request_letsencrypt?
        if letsencrypt?
          !acme.valid? || acme.expiring? || !acme.has_names?(server_names)
        end
      end

      # accessors

      def service_name
        labels['vhosts-ng/service/name'] || (raise StandardError.new("#{self.name}: name not defined"))
      end

      def auto_route
        labels['vhosts-ng/service/auto-route']
      end

      def servername
        labels['vhosts-ng/service/server-name'] || (raise StandardError.new("#{self.name}: server name not defined"))
      end

      def aliases
        (labels['vhosts-ng/service/aliases']||'').split(',').map(&:strip)
      end

      def server_names
        [auto_route, servername, aliases].flatten.reject(&:blank?)
      end

      def upstream
        labels['vhosts-ng/service/upstream'] || (raise StandardError.new("#{self.name}: upstream not defined"))
      end

      def upstream_port
        labels['vhosts-ng/service/upstream-port'] || (raise StandardError.new("#{self.name}: upstream port not defined"))
      end

      def ssl?
        labels['vhosts-ng/https'].to_s == 'true'
      end

      def http?
        labels['vhosts-ng/http'].to_s == 'true'
      end

      def letsencrypt?
        labels['vhosts-ng/letsencrypt'].to_s == 'true'
      end

      def webroot
        labels['vhosts-ng/html/webroot']|| (raise StandardError.new("#{self.name}: webroot not defined"))
      end

      def maintenance_filename
        labels['vhosts-ng/html/maintenance'] || (raise StandardError.new("#{self.name}: maintenance filename not defined"))
      end

      def unauthorized_filename
        labels['vhosts-ng/html/40x'] || (raise StandardError.new("#{self.name}: 40x filename not defined"))
      end

      def servererror_filename
        labels['vhosts-ng/html/50x'] || (raise StandardError.new("#{self.name}: 50x filename not defined"))
      end

      def max_body_size
        labels['vhosts-ng/max_body_size']
      end

      def error_log
        labels['vhosts-ng/service/error-log'] || File.join(File.dirname(webroot), 'error.log')
      end

      def access_log
        labels['vhosts-ng/service/access-log'] || (raise StandardError.new("#{self.name}: access log not defined"))
      end

      def auth_basic
        @auth_basic ||= begin
          count = 1
          auths = labels.select{|l| l =~ /vhosts-ng\/auth\/basic/}.map { | k, v |
            loc = k.gsub(/vhosts-ng\/auth\/basic/, '')
            secret, realm = v.split(',', 2)
            secret = File::expand_path(secret, '/') # sanitize....
            secret = File.join(NginxController.htpasswd_path, service_name, secret)
            unless File.exists?(secret)
              # generate a dummy htpasswd with a random username/password
              secret = File.join(NginxController.htpasswd_path, "dummy", service_name, "htpasswd.#{count}")
              FileUtils.mkdir_p(File.dirname(secret))
              File.open(secret, 'wb+') do | f |
                f.write "%s:%s" % [
                  SecureRandom.alphanumeric(8),
                  SecureRandom.alphanumeric(16).crypt("$5$#{SecureRandom.alphanumeric(8)}$"),
                ]
              end
            end
            [loc, {location: loc, secret: secret, realm: realm || 'Pretected area'}]
          }
          auths = Hash[auths]
        end
      end

      def le_validation_host
        ENV['VALIDATION_HOST']
      end

      def syslog_server
        ENV['SYSLOG_SERVER']
      end

      def letsencrypt_path
        File.join(NginxController::Acme.le_root, service_name)
      end

      def ssl_certs
        if ssl?
          @ssl_certs ||= begin
            names = [self.servername, self.aliases].flatten.uniq
            find_cert_for_names(names)
          end
        end
      end

      def auto_route_ssl_cert
        @auto_route_ssl ||= find_cert_for_names([auto_route])
        @auto_route_ssl.first.cert
      end

      def auto_route_ssl_key
        @auto_route_ssl ||= find_cert_for_names([auto_route])
        @auto_route_ssl.first.key
      end

      protected

      def create_directories
        [ File.dirname(webroot),
          File.dirname(maintenance_filename),
          File.dirname(unauthorized_filename),
          File.dirname(servererror_filename),
          File.dirname(error_log),
          File.dirname(access_log),
        ].uniq.each do | dir |
          FileUtils.mkdir_p(dir)
        end
      end

      def render_logrotate
        render 'logrotate'
      end

      def render(template = 'nginx', locals = nil)
        @locals = locals
        tpl = File.join(File.dirname(File.dirname(__FILE__)), 'templates', "#{template}.erb")
        raise StandardError.new "Template #{template} not found" unless File::exists?(tpl)
        erb = ERB.new(File.read(tpl), nil, '-')
        erb.result self.get_binding
      end

      def write_logrotate_config!(config)
        File.open(logrotate_config_filename, 'wb+') {|f| f.write(config)}
      end

      def self_signed_cert
        setup_self_signed_cert unless @self_signed_cert
        @self_signed_cert
      end

      def setup_self_signed_cert
        if ssl?
          @self_signed_cert ||= NginxController::SelfSignedCert.new({
            names: server_names,
            service_name: service_name
          })
          @self_signed_cert.create! unless @self_signed_cert.valid?
        end
      end

      def acme
        if letsencrypt?
          @acme ||= NginxController::Acme.get({
            names: server_names,
            service_name: service_name
          })
        end
      end

      def get_binding
        binding
      end

      def list_ssl_certs
        if ssl?
          @ssl_certs_list ||= begin
            list = []
            
            certs = labels.select{|k,v| k =~ /\Avhosts-ng\/ssl\/certs\/.*\Z/}
            certs.each do | id, d |
              id = File::basename(id) # LAZY but works
              names = d.split(',').map(&:strip).compact

              if names.any? && File::exists?(File.join(NginxController.certs_path, service_name, "#{id}.pem"))
                list << OpenStruct.new({
                  names: names,
                  cert: File.join(NginxController.certs_path, service_name, "#{id}.pem"),
                  key: File.join(NginxController.certs_path, service_name, "#{id}.key"),
                })
              end
            end

            if letsencrypt? && acme.valid?
              list << OpenStruct.new({
                names: acme.names,
                cert: acme.cert_path,
                key: acme.privkey_path,
              })
            end
            
            list << OpenStruct.new({
              names: self_signed_cert.names,
              cert: self_signed_cert.cert_path,
              key: self_signed_cert.privkey_path,
            })
            list
          end
        end
        @ssl_certs_list || []
      end

      # WARN, LAZY: wildcards only in (a)
      def match_domains(a,b)
        matches = a & b
        a = a - matches
        b = b - matches

        a.each do | adom |
          next unless adom =~ /\A\*\./
          re = Regexp.new "\\A#{adom.gsub(/\*/, '[^.]+')}\\Z"
          matches += b.select{|bdom| bdom =~ re }
        end
        matches
      end

      def find_cert_for_names(names)
        list_ssl_certs.inject([]) do | memo, cert |
          matches = match_domains(cert.names, names)
          if matches.any?
            names -= matches
            memo << OpenStruct.new({
              names: matches,
              cert: cert.cert,
              key: cert.key,
            })
          end
          memo
        end
      end
    end
  end
end
