
module NginxController
  class SelfSignedCert
    attr_reader :options, :pkey, :cert

    def self.cert_root
      ENV['SELF_SIGNED_SSL_ROOT'].presence || File.join(NginxController.ssl_root, 'self_signed')
    end

    def initialize(options)
      @options = options
    end

    def privkey_path
      File.join(self.class.cert_root, @options[:service_name], 'key.pem')
    end

    def cert_path
      File.join(self.class.cert_root, @options[:service_name], 'cert.pem')
    end

    def valid?
      !!pkey && !!cert && 
        names.sort == options[:names].sort &&
        expires_at > Time.now
    end

    def expires_at
      cert.try(:not_after) || Time.at(0)
    end

    def create!
      @pkey = OpenSSL::PKey::RSA.new 2048

      @cert = OpenSSL::X509::Certificate.new
      @cert.version = 2
      @cert.serial = SecureRandom.hex(10).to_i(16)
      @cert.subject = OpenSSL::X509::Name.parse("/C=IT/L=Temporary self-signed certificate/O=Temporary self-signed certificate/CN=%s" % options[:names].first)
      @cert.issuer = @cert.subject
      @cert.public_key = @pkey.public_key
      @cert.not_before = Time.now
      @cert.not_after = @cert.not_before + 2 * 365 * 24 * 60 * 60
      ef = OpenSSL::X509::ExtensionFactory.new
      ef.subject_certificate = @cert
      @cert.add_extension(ef.create_extension("subjectAltName","DNS:#{options[:names].join(',DNS:')}",false))
      @cert.sign(@pkey, OpenSSL::Digest::SHA256.new)
      
      FileUtils.mkdir_p(File.dirname(privkey_path))
      FileUtils.mkdir_p(File.dirname(cert_path))
      File.open(privkey_path, 'wb+') {|f| f.write @pkey.to_pem }
      File.open(cert_path, 'wb+') {|f| f.write @cert.to_pem }
    end

    def names
      names = []
      if cert
        name = cert.subject.to_a.find{|x| x[0] == 'CN'}
        names << name[1] if name

        ext = cert.extensions.find{|x| x.oid == 'subjectAltName'}
        names += ext.value.split(',').map{|n| n.strip.gsub(/\Adns:/i, '') } if ext
        names.uniq!
      end
      names
    end

    protected

    def load
      if File.exists?(privkey_path) && File.exists?(cert_path)
        @pkey = OpenSSL::PKey.read(File.read(privkey_path))
        @cert = OpenSSL::X509::Certificate.new(File.read(cert_path))
      end
    end
  end
end